from functools import partial
from typing import NamedTuple, List
from unittest.mock import patch
import pytest

from snakeoil.contexts import chdir
from pkgdev.scripts import run

class Profile(NamedTuple):
    """Profile record used to create profiles in a repository."""
    path: str
    arch: str
    status: str = 'stable'
    deprecated: bool = False
    defaults: List[str] = None
    eapi: str = '5'

class TestPkgdevShowkwParseArgs:

    def test_missing_target(self, capsys, tool):
        with pytest.raises(SystemExit):
            tool.parse_args(['showkw', '--config', 'no'])
        captured = capsys.readouterr()
        assert captured.err.strip() == (
            'pkgdev showkw: error: missing target argument and not in a supported repo')

    def test_unknown_arches(self, capsys, tool, make_repo):
        repo = make_repo(arches=['amd64'])
        with pytest.raises(SystemExit):
            tool.parse_args(['showkw', '--config', 'no', '-a', 'unknown', '-r', repo.location])
        captured = capsys.readouterr()
        assert captured.err.strip() == (
            "pkgdev showkw: error: unknown arch: 'unknown' (choices: amd64)")

class TestPkgdevShowkw:

    script = staticmethod(partial(run, 'pkgdev'))
    base_args = ['pkgdev', 'showkw', '--config', 'n']

    def _create_repo(self, make_repo):
        repo = make_repo(arches=['amd64', 'ia64', 'mips', 'x86'])
        repo.create_profiles([
            Profile('default/linux/amd64', 'amd64'),
            Profile('default/linux/x86', 'x86'),
            Profile('default/linux/ia64', 'ia64', 'dev'),
            Profile('default/linux/mips', 'mips', 'exp'),
        ])
        return repo

    def _run_and_parse(self, capsys, *args):
        with patch('sys.argv', [*self.base_args, "--format", "presto", *args]), \
                pytest.raises(SystemExit) as excinfo:
            self.script()
        assert excinfo.value.code == None
        out, err = capsys.readouterr()
        assert not err
        lines = out.split('\n')
        table_columns = [s.strip() for s in lines[1].split('|')][1:]
        return {
            ver: dict(zip(table_columns, values))
                for ver, *values in map(lambda s: map(str.strip, s.split('|')), lines[3:-1])
        }

    def test_match(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0')
        with patch('sys.argv', [*self.base_args, '-r', repo.location, 'foo/bar']), \
                pytest.raises(SystemExit) as excinfo:
            self.script()
        assert excinfo.value.code == None
        out, err = capsys.readouterr()
        assert not err
        assert out.split('\n')[0] == "keywords for foo/bar:"

    def test_match_short_name(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0')
        with patch('sys.argv', [*self.base_args, '-r', repo.location, 'bar']), \
                pytest.raises(SystemExit) as excinfo:
            self.script()
        assert excinfo.value.code == None
        out, err = capsys.readouterr()
        assert not err
        assert out.split('\n')[0] == "keywords for foo/bar:"

    def test_match_cwd_repo(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0')
        with patch('sys.argv', [*self.base_args, 'foo/bar']), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            self.script()
        assert excinfo.value.code == None
        out, err = capsys.readouterr()
        assert not err
        assert out.split('\n')[0] == "keywords for foo/bar:"

    def test_match_cwd_pkg(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0')
        with patch('sys.argv', self.base_args), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location + '/foo/bar'):
            self.script()
        assert excinfo.value.code == None
        _, err = capsys.readouterr()
        assert not err

    def test_no_matches(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        with patch('sys.argv', [*self.base_args, '-r', repo.location, 'foo/bar']), \
                pytest.raises(SystemExit) as excinfo:
            self.script()
        assert excinfo.value.code == 1
        out, err = capsys.readouterr()
        assert not out
        assert err.strip() == "pkgdev showkw: no matches for 'foo/bar'"

    def test_match_stable(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0', keywords=('~amd64', '~ia64', '~mips', 'x86'))
        res = self._run_and_parse(capsys, '-r', repo.location, 'foo/bar', '--stable')
        assert set(res.keys()) == {'0'}
        assert {'amd64', 'ia64', 'mips', 'x86'} & res['0'].keys() == {'amd64', 'x86'}

    def test_match_unstable(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0', keywords=('~amd64', '~ia64', '~mips', 'x86'))
        res = self._run_and_parse(capsys, '-r', repo.location, 'foo/bar', '--unstable')
        assert set(res.keys()) == {'0'}
        assert {'amd64', 'ia64', 'mips', 'x86'} <= res['0'].keys()

    def test_match_specific_arch(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0', keywords=('~amd64', '~ia64', '~mips', 'x86'))
        res = self._run_and_parse(capsys, '-r', repo.location, 'foo/bar', '--arch', 'amd64')
        assert set(res.keys()) == {'0'}
        assert {'amd64', 'ia64', 'mips', 'x86'} & res['0'].keys() == {'amd64'}

    def test_match_specific_multiple_arch(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0', keywords=('~amd64', '~ia64', '~mips', 'x86'))
        res = self._run_and_parse(capsys, '-r', repo.location, 'foo/bar', '--arch', 'amd64,mips')
        assert set(res.keys()) == {'0'}
        assert {'amd64', 'ia64', 'mips', 'x86'} & res['0'].keys() == {'amd64', 'mips'}

    def test_correct_keywords_status(self, capsys, make_repo):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0', keywords=('amd64', '~ia64', '~mips', 'x86'))
        repo.create_ebuild('foo/bar-1', keywords=('~amd64', '-mips', '~x86'))
        repo.create_ebuild('foo/bar-2', keywords=('-*', 'amd64', '-x86'), eapi=8, slot=2)
        res = self._run_and_parse(capsys, '-r', repo.location, 'foo/bar')
        assert set(res.keys()) == {'0', '1', '2'}
        assert dict(amd64='+', ia64='~', mips='~', x86='+', slot='0').items() <= res['0'].items()
        assert dict(amd64='~', ia64='o', mips='-', x86='~', slot='0').items() <= res['1'].items()
        assert dict(amd64='+', ia64='*', mips='*', x86='-', slot='2', eapi='8').items() <= res['2'].items()

    @pytest.mark.parametrize(('arg', 'expected'),
        (
            ('--stable', {'amd64', 'x86'}),
            ('--unstable', {'amd64', 'ia64', 'mips', 'x86'}),
            ('--only-unstable', {'ia64', 'mips'}),
        )
    )
    def test_collapse(self, capsys, make_repo, arg, expected):
        repo = self._create_repo(make_repo)
        repo.create_ebuild('foo/bar-0', keywords=('amd64', '~ia64', '~mips', '~x86'))
        repo.create_ebuild('foo/bar-1', keywords=('~amd64', '~ia64', '~mips', 'x86'))
        with patch('sys.argv', [*self.base_args, '-r', repo.location, 'foo/bar', "--collapse", arg]), \
                pytest.raises(SystemExit) as excinfo:
            self.script()
        out, err = capsys.readouterr()
        assert excinfo.value.code == None
        assert not err
        arches = set(out.split('\n')[0].split())
        assert arches == expected



from functools import partial
from typing import List, Set
from unittest.mock import patch

import pytest
from pkgdev.scripts import run
from snakeoil.contexts import chdir
from snakeoil.osutils import pjoin


class TestPkgdevManifestParseArgs:

    def test_non_repo_cwd(self, capsys, tool):
        with pytest.raises(SystemExit) as excinfo:
            tool.parse_args(['manifest'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev manifest: error: not in ebuild repo'

    def test_repo_cwd(self, repo, capsys, tool):
        repo.create_ebuild('cat/pkg-0')
        with chdir(repo.location):
            options, _ = tool.parse_args(['manifest'])
        matches = [x.cpvstr for x in repo.itermatch(options.restriction)]
        assert matches == ['cat/pkg-0']

    def test_repo_relative_pkg(self, repo, capsys, tool):
        repo.create_ebuild('cat/pkg-0')
        repo.create_ebuild('cat/newpkg-0')
        with chdir(pjoin(repo.location, 'cat/pkg')):
            options, _ = tool.parse_args(['manifest', '.'])
        matches = [x.cpvstr for x in repo.itermatch(options.restriction)]
        assert matches == ['cat/pkg-0']

    def test_repo_relative_category(self, repo, capsys, tool):
        repo.create_ebuild('cat/pkg-0')
        repo.create_ebuild('cat/newpkg-0')

        with chdir(pjoin(repo.location, 'cat')):
            options, _ = tool.parse_args(['manifest', 'pkg'])
        matches = [x.cpvstr for x in repo.itermatch(options.restriction)]
        assert matches == ['cat/pkg-0']

        with chdir(pjoin(repo.location, 'cat')):
            options, _ = tool.parse_args(['manifest', '.'])
        matches = [x.cpvstr for x in repo.itermatch(options.restriction)]
        assert set(matches) == {'cat/pkg-0', 'cat/newpkg-0'}

    def test_repo_relative_outside(self, tmp_path, repo, capsys, tool):
        repo.create_ebuild('cat/pkg-0')
        (ebuild := tmp_path / 'pkg.ebuild').touch()
        with pytest.raises(SystemExit) as excinfo:
            with chdir(repo.location):
                tool.parse_args(['manifest', str(ebuild)])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.strip() == f"pkgdev manifest: error: {repo.repo_id!r} repo doesn't contain: {str(ebuild)!r}"

    def test_dir_target(self, repo, capsys, tool):
        repo.create_ebuild('cat/pkg-0')
        with chdir(repo.location):
            options, _ = tool.parse_args(['manifest', pjoin(repo.location, 'cat')])
        matches = [x.cpvstr for x in repo.itermatch(options.restriction)]
        assert matches == ['cat/pkg-0']

    def test_ebuild_target(self, repo, capsys, tool):
        path = repo.create_ebuild('cat/pkg-0')
        with chdir(repo.location):
            options, _ = tool.parse_args(['manifest', path])
        matches = [x.cpvstr for x in repo.itermatch(options.restriction)]
        assert matches == ['cat/pkg-0']

    def test_atom_target(self, repo, capsys, tool):
        repo.create_ebuild('cat/pkg-0')
        with chdir(repo.location):
            options, _ = tool.parse_args(['manifest', 'cat/pkg'])
        matches = [x.cpvstr for x in repo.itermatch(options.restriction)]
        assert matches == ['cat/pkg-0']

    def test_if_modified_target(self, repo, make_git_repo, tool):
        def manifest_matches() -> Set[str]:
            repo.sync()
            with chdir(repo.location):
                options, _ = tool.parse_args(['manifest', '--if-modified'])
            return {x.cpvstr for x in repo.itermatch(options.restriction)}

        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/oldpkg-0')
        git_repo.add_all('cat/oldpkg-0')

        # New package
        repo.create_ebuild('cat/newpkg-0')
        assert manifest_matches() == {'cat/newpkg-0'}
        git_repo.add_all('cat/newpkg-0')

        # Untracked file
        ebuild_path = repo.create_ebuild('cat/newpkg-1')
        assert manifest_matches() == {'cat/newpkg-1'}

        # Staged file
        git_repo.add(ebuild_path, commit=False)
        assert manifest_matches() == {'cat/newpkg-1'}

        # No modified files
        git_repo.add_all('cat/newpkg-1')
        assert manifest_matches() == set()

        # Modified file
        ebuild_path = repo.create_ebuild('cat/newpkg-1', eapi=8)
        assert manifest_matches() == {'cat/newpkg-1'}
        git_repo.add_all('cat/newpkg-1: eapi 8')

        # Renamed file
        git_repo.remove(ebuild_path, commit=False)
        ebuild_path = repo.create_ebuild('cat/newpkg-2')
        git_repo.add(ebuild_path, commit=False)
        assert manifest_matches() == {'cat/newpkg-2'}
        git_repo.add_all('cat/newpkg-2: rename')

        # Deleted file
        git_repo.remove(ebuild_path, commit=False)
        assert manifest_matches() == set()

        # Deleted package
        ebuild_path = repo.create_ebuild('cat/newpkg-0')
        git_repo.remove(ebuild_path, commit=False)
        assert manifest_matches() == set()

    def test_ignore_fetch_restricted(self, repo, tool):
        def manifest_matches() -> List[str]:
            with chdir(repo.location):
                options, _ = tool.parse_args(['manifest', '--ignore-fetch-restricted'])
            return [x.cpvstr for x in repo.itermatch(options.restriction)]

        # No RESTRICT
        repo.create_ebuild('cat/pkg-0')
        assert manifest_matches() == ['cat/pkg-0']

        # Not fetch RESTRICT
        repo.create_ebuild('cat/pkg-0', restrict=('mirror'))
        assert manifest_matches() == ['cat/pkg-0']

        # fetch RESTRICT
        repo.create_ebuild('cat/pkg-0', restrict=('fetch'))
        assert manifest_matches() == []

        # Multiple RESTRICT
        repo.create_ebuild('cat/pkg-0', restrict=('mirror', 'fetch'))
        assert manifest_matches() == []

    def test_non_repo_dir_target(self, tmp_path, repo, capsys, tool):
        with pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            tool.parse_args(['manifest', str(tmp_path)])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.startswith("pkgdev manifest: error: 'fake' repo doesn't contain:")

    def test_invalid_atom_target(self, repo, capsys, tool):
        with pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            tool.parse_args(['manifest', '=cat/pkg'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.startswith("pkgdev manifest: error: invalid atom: '=cat/pkg'")


class TestPkgdevManifest:

    script = staticmethod(partial(run, 'pkgdev'))

    @pytest.fixture(autouse=True)
    def _setup(self):
        self.args = ['pkgdev', 'manifest']

    def test_good_manifest(self, capsys, repo):
        repo.create_ebuild('cat/pkg-0')
        with patch('sys.argv', self.args), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            self.script()
        assert excinfo.value.code == 0
        out, err = capsys.readouterr()
        assert out == err == ''

    def test_bad_manifest(self, capsys, repo):
        repo.create_ebuild('cat/pkg-0')
        repo.create_ebuild('cat/pkg-1', eapi='-1')
        with patch('sys.argv', self.args), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            self.script()
        assert excinfo.value.code == 1
        out, err = capsys.readouterr()
        assert not err
        assert out == " * cat/pkg-1: invalid EAPI '-1'\n"



"""Formatting and file mangling support."""

import functools
import multiprocessing
import os
import re
import signal
import traceback
from datetime import datetime

from snakeoil.cli.exceptions import UserException
from snakeoil.mappings import OrderedSet

copyright_regex = re.compile(
    r'^# Copyright (?P<date>(?P<begin>\d{4}-)?(?P<end>\d{4})) (?P<holder>.+)$')

keywords_regex = re.compile(
    r'^(?P<pre>[^#]*\bKEYWORDS=(?P<quote>[\'"]?))'
    r'(?P<keywords>.*)'
    r'(?P<post>(?P=quote).*)$')


def mangle(name):
    """Decorator to register file mangling methods."""

    class decorator:
        """Decorator with access to the class of a decorated function."""

        def __init__(self, func):
            self.func = func

        def __set_name__(self, owner, name):
            owner._mangle_funcs[name] = self.func
            setattr(owner, name, self.func)

    return decorator


class Mangler:
    """File-mangling iterator using path-based parallelism."""

    # mapping of mangling types to functions
    _mangle_funcs = {}

    def __init__(self, changes, skip_regex=None):
        self.jobs = os.cpu_count()
        if skip_regex is not None:
            changes = (c for c in changes if not skip_regex.match(c.full_path))
        self.changes = OrderedSet(changes)

        # setup for parallelizing the mangling procedure across files
        self._mp_ctx = multiprocessing.get_context('fork')
        self._mangled_paths_q = self._mp_ctx.SimpleQueue()
        self._current_year = str(datetime.today().year)

        # initialize settings used by iterator support
        self._runner = self._mp_ctx.Process(target=self._run)
        signal.signal(signal.SIGINT, self._kill_pipe)
        self._mangled_paths = iter(self._mangled_paths_q.get, None)

        # construct composed mangling function
        self.composed_func = functools.reduce(
            lambda f, g: lambda x: f(g(self, x)), self._mangle_funcs.values(), lambda x: x)

    @mangle('EOF')
    def _eof(self, change):
        """Drop EOF whitespace and forcibly add EOF newline."""
        return change.update(change.data.rstrip() + '\n')

    @mangle('keywords')
    def _keywords(self, change):
        """Fix keywords order."""

        def keywords_sort_key(kw):
            return tuple(reversed(kw.lstrip('-~').partition('-')))

        lines = change.data.splitlines()
        for i, line in enumerate(lines):
            if mo := keywords_regex.match(line):
                kw = sorted(mo.group('keywords').split(), key=keywords_sort_key)
                new_kw = ' '.join(kw)
                if not mo.group('quote'):
                    new_kw = f'"{new_kw}"'
                lines[i] = f'{mo.group("pre")}{new_kw}{mo.group("post")}'
                break
        return change.update('\n'.join(lines) + '\n')

    def _kill_pipe(self, *args, error=None):
        """Handle terminating the mangling process group."""
        if self._runner.is_alive():
            os.killpg(self._runner.pid, signal.SIGKILL)
        if error is not None:
            # propagate exception raised during parallelized mangling
            raise UserException(error)
        raise KeyboardInterrupt

    def __iter__(self):
        # start running the mangling processes
        self._runner.start()
        return self

    def __next__(self):
        try:
            path = next(self._mangled_paths)
        except StopIteration:
            self._runner.join()
            raise

        # Catch propagated, serialized exceptions, output their
        # traceback, and signal the scanning process to end.
        if isinstance(path, list):
            self._kill_pipe(error=path[0])

        return path

    def _mangle(self, change):
        """Run composed mangling function across a given change."""
        if orig_data := change.read():
            change = self.composed_func(change)
            if change.data != orig_data:
                change.sync()
                return change

    def _run_manglers(self, paths_q):
        """Consumer that runs mangling functions, queuing mangled paths for output."""
        try:
            for change in iter(paths_q.get, None):
                if mangled_change := self._mangle(change):
                    self._mangled_paths_q.put(mangled_change.path)
        except Exception:  # pragma: no cover
            # traceback can't be pickled so serialize it
            tb = traceback.format_exc()
            self._mangled_paths_q.put([tb])

    def _run(self):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        os.setpgrp()

        paths_q = self._mp_ctx.SimpleQueue()
        pool = self._mp_ctx.Pool(self.jobs, self._run_manglers, (paths_q,))
        pool.close()

        # queue paths for processing
        for change in self.changes:
            paths_q.put(change)
        # notify consumers that no more work exists
        for i in range(self.jobs):
            paths_q.put(None)

        pool.join()
        # notify iterator that no more results exist
        self._mangled_paths_q.put(None)


class GentooMangler(Mangler):
    """Gentoo repo specific file mangler."""

    _mangle_funcs = Mangler._mangle_funcs.copy()

    @mangle('copyright')
    def _copyright(self, change):
        """Fix copyright headers and dates."""
        lines = change.data.splitlines()
        if mo := copyright_regex.match(lines[0]):
            groups = mo.groupdict()
            if groups['begin'] is None and groups['date'] != self._current_year:
                # use old copyright date as the start of date range
                date_range = f"{groups['date']}-{self._current_year}"
                lines[0] = re.sub(groups['date'], date_range, lines[0])
            else:
                lines[0] = re.sub(mo.group('end'), self._current_year, lines[0])
            lines[0] = re.sub('Gentoo Foundation', 'Gentoo Authors', lines[0])
        return change.update('\n'.join(lines) + '\n')



import os
import re
import subprocess

from pkgcore.operations import observer as observer_mod
from pkgcore.restrictions import packages, values
from pkgcore.util.parserestrict import parse_match
from snakeoil.cli import arghparse

from .. import cli, git
from .argparsers import cwd_repo_argparser

manifest = cli.ArgumentParser(
    prog='pkgdev manifest', description='update package manifests',
    parents=(cwd_repo_argparser,))
manifest.add_argument(
    'target', nargs='*',
    help='packages to target',
    docs="""
        Packages matching any of these restrictions will have their manifest
        entries updated. If no target is specified a path restriction is
        created based on the current working directory. In other words, if
        ``pkgdev manifest`` is run within an ebuild's directory, all the
        ebuilds within that directory will be manifested.
    """)
manifest_opts = manifest.add_argument_group('manifest options')
manifest_opts.add_argument(
    '-d', '--distdir', type=arghparse.create_dir, help='target download directory',
    docs="""
        Use a specified target directory for downloads instead of the
        configured DISTDIR.
    """)
manifest_opts.add_argument(
    '-f', '--force', help='forcibly remanifest packages',
    action='store_true',
    docs="""
        Force package manifest files to be rewritten. Note that this requires
        downloading all distfiles.
    """)
manifest_opts.add_argument(
    '-m', '--mirrors', help='enable fetching from Gentoo mirrors',
    action='store_true',
    docs="""
        Enable checking Gentoo mirrors first for distfiles. This is disabled by
        default because manifest generation is often performed when adding new
        ebuilds with distfiles that aren't on Gentoo mirrors yet.
    """)
manifest_opts.add_argument(
    '--if-modified', dest='if_modified', help='Only check packages that have uncommitted modifications',
    action='store_true',
    docs="""
        In addition to matching the specified restriction, restrict to targets
        which are marked as modified by git, including untracked files.
    """)
manifest_opts.add_argument(
    '--ignore-fetch-restricted', dest='ignore_fetch_restricted', help='Ignore fetch restricted ebuilds',
    action='store_true',
    docs="""
        Ignore attempting to update manifest entries for ebuilds which are
        fetch restricted.
    """)


def _restrict_targets(repo, targets):
    restrictions = []
    for target in targets:
        if os.path.exists(target):
            try:
                if target in repo:
                    target = os.path.relpath(target, repo.location)
                restrictions.append(repo.path_restrict(target))
            except ValueError as exc:
                manifest.error(exc)
        else:
            try:
                restrictions.append(parse_match(target))
            except ValueError:
                manifest.error(f'invalid atom: {target!r}')
    return packages.OrRestriction(*restrictions)


def _restrict_modified_files(repo):
    ebuild_re = re.compile(r'^[ MTARC?]{2} (?P<path>[^/]+/[^/]+/[^/]+\.ebuild)$')
    p = git.run('status', '--porcelain=v1', '-z', "*.ebuild",
                cwd=repo.location, stdout=subprocess.PIPE)

    restrictions = []
    for line in p.stdout.strip('\x00').split('\x00'):
        if mo := ebuild_re.match(line):
            restrictions.append(repo.path_restrict(mo.group('path')))
    return packages.OrRestriction(*restrictions)


@manifest.bind_final_check
def _manifest_validate(parser, namespace):
    targets = namespace.target if namespace.target else [namespace.cwd]

    restrictions = [_restrict_targets(namespace.repo, targets)]
    if namespace.if_modified:
        restrictions.append(_restrict_modified_files(namespace.repo))
    if namespace.ignore_fetch_restricted:
        restrictions.append(packages.PackageRestriction('restrict', values.ContainmentMatch('fetch', negate=True)))
    namespace.restriction = packages.AndRestriction(*restrictions)


@manifest.bind_main_func
def _manifest(options, out, err):
    failed = options.repo.operations.manifest(
        domain=options.domain,
        restriction=options.restriction,
        observer=observer_mod.formatter_output(out),
        mirrors=options.mirrors,
        force=options.force,
        distdir=options.distdir)

    return int(any(failed))



import argparse
import shlex

from pkgcheck import reporters, scan
from snakeoil.cli import arghparse
from snakeoil.cli.input import userquery

from .. import cli, git
from .argparsers import cwd_repo_argparser, git_repo_argparser


class ArgumentParser(cli.ArgumentParser):
    """Parse all known arguments, passing unknown arguments to ``git push``."""

    def parse_known_args(self, args=None, namespace=None):
        namespace, args = super().parse_known_args(args, namespace)
        if namespace.dry_run:
            args.append('--dry-run')
        namespace.push_args = args
        return namespace, []


push = ArgumentParser(
    prog='pkgdev push', description='run QA checks on commits and push them',
    parents=(cwd_repo_argparser, git_repo_argparser))
# custom `pkgcheck scan` args used for tests
push.add_argument('--pkgcheck-scan', help=argparse.SUPPRESS)
push_opts = push.add_argument_group('push options')
push_opts.add_argument(
    '-A', '--ask', nargs='?', const=True, action=arghparse.StoreBool,
    help='confirm pushing commits with QA errors')
push_opts.add_argument(
    '-n', '--dry-run', action='store_true',
    help='pretend to push the commits')


@push.bind_final_check
def _commit_validate(parser, namespace):
    # determine `pkgcheck scan` args
    namespace.scan_args = ['-v'] * namespace.verbosity
    if namespace.pkgcheck_scan:
        namespace.scan_args.extend(shlex.split(namespace.pkgcheck_scan))
    namespace.scan_args.extend(['--exit', 'GentooCI', '--commits'])


@push.bind_main_func
def _push(options, out, err):
    # scan commits for QA issues
    pipe = scan(options.scan_args)
    with reporters.FancyReporter(out) as reporter:
        for result in pipe:
            reporter.report(result)

    # fail on errors unless they're ignored
    if pipe.errors:
        with reporters.FancyReporter(out) as reporter:
            out.write(out.bold, out.fg('red'), '\nFAILURES', out.reset)
            for result in sorted(pipe.errors):
                reporter.report(result)
        if not (options.ask and userquery('Push commits anyway?', out, err)):
            return 1

    # push commits upstream
    git.run('push', *options.push_args, cwd=options.repo.location)

    return 0



import os
import subprocess

from pkgcore.repository import errors as repo_errors
from snakeoil.cli.arghparse import ArgumentParser

from .. import git

cwd_repo_argparser = ArgumentParser(suppress=True)
git_repo_argparser = ArgumentParser(suppress=True)


@cwd_repo_argparser.bind_final_check
def _determine_cwd_repo(parser, namespace):
    namespace.cwd = os.getcwd()
    try:
        repo = namespace.domain.find_repo(
            namespace.cwd, config=namespace.config, configure=False)
    except (repo_errors.InitializationError, IOError) as e:
        raise parser.error(str(e))

    if repo is None:
        raise parser.error('not in ebuild repo')

    namespace.repo = repo


@git_repo_argparser.bind_final_check
def _determine_git_repo(parser, namespace):
    try:
        p = git.run('rev-parse', '--show-toplevel', stdout=subprocess.PIPE)
        path = p.stdout.strip()
    except git.GitError:
        raise parser.error('not in git repo')

    # verify the git and ebuild repo roots match when using both
    try:
        if namespace.repo.location != path:
            raise parser.error('not in ebuild git repo')
    except AttributeError:
        # ebuild repo parser not enabled
        pass

    namespace.git_repo = path



import pytest
from pkgdev.cli import Tool
from pkgdev.scripts import pkgdev
from snakeoil.cli import arghparse

pytest_plugins = ['pkgcore']


@pytest.fixture(scope="session")
def tool():
    """Generate a tool utility for running pkgdev."""
    return Tool(pkgdev.argparser)


@pytest.fixture
def parser():
    """Return a shallow copy of the main pkgdev argparser."""
    return pkgdev.argparser.copy()


@pytest.fixture
def namespace():
    """Return an arghparse Namespace object."""
    return arghparse.Namespace()



import importlib
from functools import partial
from unittest.mock import patch

import pytest

from pkgdev import __title__ as project
from pkgdev.scripts import run


def test_script_run(capsys):
    """Test regular code path for running scripts."""
    script = partial(run, project)

    with patch(f'{project}.scripts.import_module') as import_module:
        import_module.side_effect = ImportError("baz module doesn't exist")

        # default error path when script import fails
        with patch('sys.argv', [project]):
            with pytest.raises(SystemExit) as excinfo:
                script()
            assert excinfo.value.code == 1
            out, err = capsys.readouterr()
            err = err.strip().split('\n')
            assert len(err) == 3
            assert err[0] == "Failed importing: baz module doesn't exist!"
            assert err[1].startswith(f"Verify that {project} and its deps")
            assert err[2] == "Add --debug to the commandline for a traceback."

        # running with --debug should raise an ImportError when there are issues
        with patch('sys.argv', [project, '--debug']):
            with pytest.raises(ImportError):
                script()
            out, err = capsys.readouterr()
            err = err.strip().split('\n')
            assert len(err) == 2
            assert err[0] == "Failed importing: baz module doesn't exist!"
            assert err[1].startswith(f"Verify that {project} and its deps")

        import_module.reset_mock()


class TestPkgdev:

    script = staticmethod(partial(run, project))

    def test_version(self, capsys):
        with patch('sys.argv', [project, '--version']):
            with pytest.raises(SystemExit) as excinfo:
                self.script()
            assert excinfo.value.code == 0
            out, err = capsys.readouterr()
            assert out.startswith(project)

    def test_installed(self):
        """Verify tests are running in environment where generated modules exist."""
        try:
            importlib.import_module(f'{project}._verinfo')
        except ImportError:
            pytest.fail(
                'not running against installed or released package\n'
                '(use `setup.py test` when running from git)'
            )



import os
import sys
import textwrap
from datetime import datetime, timedelta, timezone
from functools import partial
from pathlib import Path
from unittest.mock import patch

import pytest
from pkgcore.ebuild.atom import atom as atom_cls
from pkgdev.scripts import run
from snakeoil.contexts import chdir, os_environ
from snakeoil.osutils import pjoin


class TestPkgdevMaskParseArgs:

    def test_non_repo_cwd(self, capsys, tool):
        with pytest.raises(SystemExit):
            tool.parse_args(['mask'])
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev mask: error: not in ebuild repo'

    def test_non_git_repo_cwd(self, repo, capsys, tool):
        with pytest.raises(SystemExit), \
                chdir(repo.location):
            tool.parse_args(['mask'])
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev mask: error: not in git repo'

    def test_non_ebuild_git_repo_cwd(self, make_repo, git_repo, capsys, tool):
        os.mkdir(pjoin(git_repo.path, 'repo'))
        repo = make_repo(pjoin(git_repo.path, 'repo'))
        with pytest.raises(SystemExit), \
                chdir(repo.location):
            tool.parse_args(['mask'])
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev mask: error: not in ebuild git repo'

    def test_cwd_target(self, repo, make_git_repo, capsys, tool):
        git_repo = make_git_repo(repo.location)
        # empty repo
        with pytest.raises(SystemExit), \
                chdir(repo.location):
            tool.parse_args(['mask'])
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev mask: error: not in a package directory'

        # not in package dir
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        with pytest.raises(SystemExit), \
                chdir(repo.location):
            tool.parse_args(['mask'])
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev mask: error: not in a package directory'

        # masking CWD package
        with chdir(pjoin(repo.location, 'cat/pkg')):
            options, _ = tool.parse_args(['mask'])
        assert options.atoms == [atom_cls('cat/pkg')]

    def test_targets(self, repo, make_git_repo, capsys, tool):
        git_repo = make_git_repo(repo.location)

        # invalid atom
        with pytest.raises(SystemExit), \
                chdir(repo.location):
            tool.parse_args(['mask', 'pkg'])
        out, err = capsys.readouterr()
        assert err.strip() == "pkgdev mask: error: invalid atom: 'pkg'"

        # nonexistent pkg
        with pytest.raises(SystemExit), \
                chdir(repo.location):
            tool.parse_args(['mask', 'cat/nonexistent'])
        out, err = capsys.readouterr()
        assert err.strip() == "pkgdev mask: error: no repo matches: 'cat/nonexistent'"

        # masked pkg
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        with chdir(repo.location):
            options, _ = tool.parse_args(['mask', 'cat/pkg'])
        assert options.atoms == [atom_cls('cat/pkg')]

    def test_email_not_rites(self, repo, make_git_repo, capsys, tool):
        git_repo = make_git_repo(repo.location)

        # masked pkg
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        with pytest.raises(SystemExit), \
                chdir(repo.location):
            tool.parse_args(['mask', '--email', 'cat/pkg'])
        _, err = capsys.readouterr()
        assert err.strip() == "pkgdev mask: error: last rites required for email support"


class TestPkgdevMask:

    script = staticmethod(partial(run, 'pkgdev'))

    @pytest.fixture(autouse=True)
    def _setup(self, make_repo, make_git_repo):
        # args for running pkgdev like a script
        self.args = ['pkgdev', 'mask']
        self.repo = make_repo(arches=['amd64'])
        self.git_repo = make_git_repo(self.repo.location)
        self.today = datetime.now(timezone.utc)

        # add stub pkg
        self.repo.create_ebuild('cat/pkg-0')
        self.git_repo.add_all('cat/pkg-0')

        # create profile
        self.profile_path = pjoin(self.repo.location, 'profiles/arch/amd64')
        os.makedirs(self.profile_path)
        with open(pjoin(self.repo.location, 'profiles/profiles.desc'), 'w') as f:
            f.write('amd64 arch/amd64 stable\n')

        self.masks_path = Path(pjoin(self.repo.location, 'profiles/package.mask'))

    @property
    def profile(self):
        profile = list(self.repo.config.profiles)[0]
        return self.repo.config.profiles.create_profile(profile)

    def test_empty_repo(self):
        assert self.profile.masks == frozenset()

    def test_nonexistent_editor(self, capsys):
        with os_environ('VISUAL', EDITOR='12345'), \
                patch('sys.argv', self.args + ['cat/pkg']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        out, err = capsys.readouterr()
        assert err.strip() == "pkgdev mask: error: nonexistent editor: '12345'"

    def test_nonexistent_visual(self, capsys):
        with os_environ('EDITOR', VISUAL='12345'), \
                patch('sys.argv', self.args + ['cat/pkg']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        out, err = capsys.readouterr()
        assert err.strip() == "pkgdev mask: error: nonexistent editor: '12345'"

    def test_failed_editor(self, capsys):
        with os_environ('VISUAL', EDITOR="sed -i 's///'"), \
                patch('sys.argv', self.args + ['cat/pkg']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        out, err = capsys.readouterr()
        assert err.strip() == "pkgdev mask: error: failed writing mask comment"

    def test_empty_mask_comment(self, capsys):
        with os_environ('VISUAL', EDITOR="sed -i 's/#/#/'"), \
                patch('sys.argv', self.args + ['cat/pkg']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        out, err = capsys.readouterr()
        assert err.strip() == "pkgdev mask: error: empty mask comment"

    def test_mask_cwd(self):
        with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                patch('sys.argv', self.args), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path, 'cat/pkg')):
            self.script()
        assert self.profile.masks == frozenset([atom_cls('cat/pkg')])

    def test_mask_target(self):
        with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                patch('sys.argv', self.args + ['cat/pkg']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        assert self.profile.masks == frozenset([atom_cls('cat/pkg')])

    def test_mask_ebuild_path(self):
        with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                patch('sys.argv', self.args + ['cat/pkg/pkg-0.ebuild']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        assert self.profile.masks == frozenset([atom_cls('=cat/pkg-0')])

    def test_existing_masks(self):
        self.masks_path.write_text(textwrap.dedent("""\
            # Random Dev <random.dev@email.com> (2021-03-24)
            # masked
            cat/masked
        """))

        with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                patch('sys.argv', self.args + ['=cat/pkg-0']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        assert self.profile.masks == frozenset([atom_cls('cat/masked'), atom_cls('=cat/pkg-0')])

    def test_invalid_header(self, capsys):
        self.masks_path.write_text(textwrap.dedent("""\
            # Random Dev <random.dev@email.com> (2022-09-09)
            #
            # Larry the Cow was here
            #
            # masked
            cat/masked

            # Larry the Cow <larry@gentoo.org> (2022-09-09)
            #test
            # Larry the Cow wasn't here
            cat/masked2
        """))

        with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                patch('sys.argv', self.args + ['=cat/pkg-0']), \
                pytest.raises(SystemExit), \
                chdir(pjoin(self.repo.path)):
            self.script()
        _, err = capsys.readouterr()
        assert 'invalid mask entry header, lineno 9' in err

    def test_invalid_author(self, capsys):
        for line in (
            '# Random Dev <random.dev@email.com>',
            '# Random Dev <random.dev@email.com) (2021-03-24)',
            '# Random Dev (2021-03-24)',
            '# Random Dev <random.dev@email.com> 2021-03-24',
            '# Random Dev <random.dev@email.com> (24-03-2021)',
        ):
            self.masks_path.write_text(textwrap.dedent(f"""\
                # Random Dev <random.dev@email.com> (2021-03-24)
                # masked
                cat/masked

                {line}
                # masked
                cat/masked2
            """))

            with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                    patch('sys.argv', self.args + ['=cat/pkg-0']), \
                    pytest.raises(SystemExit), \
                    chdir(pjoin(self.repo.path)):
                self.script()
            _, err = capsys.readouterr()
            assert 'pkgdev mask: error: invalid author, lineno 5' in err

    def test_last_rites(self):
        for rflag in ('-r', '--rites'):
            for args in ([rflag], [rflag, '14']):
                with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                        patch('sys.argv', self.args + ['cat/pkg'] + args), \
                        pytest.raises(SystemExit), \
                        chdir(pjoin(self.repo.path)):
                    self.script()

                days = 30 if len(args) == 1 else int(args[1])
                removal_date = self.today + timedelta(days=days)
                today = self.today.strftime('%Y-%m-%d')
                removal = removal_date.strftime('%Y-%m-%d')
                assert self.masks_path.read_text() == textwrap.dedent(f"""\
                    # First Last <first.last@email.com> ({today})
                    # mask comment
                    # Removal: {removal}.
                    cat/pkg
                """)
                self.masks_path.write_text("")  # Reset the contents of package.mask

    @pytest.mark.skipif(sys.platform == "darwin", reason="no xdg-email on mac os")
    def test_last_rites_with_email(self, tmp_path):
        output_file = tmp_path / 'mail.txt'
        for rflag in ('-r', '--rites'):
            with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'", MAILER=f"> {output_file} echo"), \
                    patch('sys.argv', self.args + ['cat/pkg', rflag, '--email']), \
                    pytest.raises(SystemExit), \
                    chdir(pjoin(self.repo.path)):
                self.script()
            out = output_file.read_text()
            assert 'mailto:gentoo-dev-announce@lists.gentoo.org' in out

            self.masks_path.write_text("")  # Reset the contents of package.mask

    @pytest.mark.skipif(sys.platform == "darwin", reason="no xdg-email on mac os")
    def test_last_email_bad_mailer(self, capsys):
        for rflag in ('-r', '--rites'):
            with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'", MAILER="false"), \
                    patch('sys.argv', self.args + ['cat/pkg', rflag, '--email']), \
                    pytest.raises(SystemExit), \
                    chdir(pjoin(self.repo.path)):
                self.script()
            _, err = capsys.readouterr()
            assert err.strip() == "pkgdev mask: error: failed opening email composer"

    def test_mask_bugs(self):
        today = self.today.strftime('%Y-%m-%d')
        for bflag in ('-b', '--bug'):
            for (bug_nums, expected) in [
                (['42'], 'Bug #42.'),
                (['42', '43'], 'Bugs #42, #43.'),
            ]:
                args = []
                for bug_num in bug_nums:
                    args += [bflag, bug_num]
                with os_environ('VISUAL', EDITOR="sed -i '1s/$/mask comment/'"), \
                        patch('sys.argv', self.args + ['cat/pkg'] + args), \
                        pytest.raises(SystemExit), \
                        chdir(pjoin(self.repo.path)):
                    self.script()

                assert self.masks_path.read_text() == textwrap.dedent(f"""\
                    # First Last <first.last@email.com> ({today})
                    # mask comment
                    # {expected}
                    cat/pkg
                """)
                self.masks_path.write_text("")  # Reset the contents of package.mask

    def test_mask_bug_bad(self, capsys, tool):
        for (arg, expected) in [('-1', 'must be >= 1'), ('foo', 'invalid integer value')]:
            with pytest.raises(SystemExit):
                tool.parse_args(['mask', '--bug', arg])
            out, err = capsys.readouterr()
            assert err.strip() == f'pkgdev mask: error: argument -b/--bug: {expected}'



"""Internal constants."""

import os
import sys

from snakeoil import mappings

_reporoot = os.path.realpath(__file__).rsplit(os.path.sep, 3)[0]
_module = sys.modules[__name__]

try:
    # This is a file written during installation;
    # if it exists, we defer to it. If it doesn't, then we're
    # running from a git checkout or a tarball.
    from . import _const as _defaults
except ImportError:  # pragma: no cover
    _defaults = object()


def _GET_CONST(attr, default_value):
    consts = mappings.ProxiedAttrs(_module)
    default_value %= consts
    return getattr(_defaults, attr, default_value)


# determine XDG compatible paths
for xdg_var, var_name, fallback_dir in (
        ('XDG_CONFIG_HOME', 'USER_CONFIG_PATH', '~/.config'),
        ('XDG_DATA_HOME', 'USER_DATA_PATH', '~/.local/share')):
    setattr(
        _module, var_name,
        os.path.join(os.environ.get(xdg_var, os.path.expanduser(fallback_dir)), 'pkgdev'))

REPO_PATH = _GET_CONST('REPO_PATH', _reporoot)
DATA_PATH = _GET_CONST('DATA_PATH', '%(REPO_PATH)s/data')

USER_CONF_FILE = os.path.join(getattr(_module, 'USER_CONFIG_PATH'), 'pkgdev.conf')
SYSTEM_CONF_FILE = '/etc/pkgdev/pkgdev.conf'
BUNDLED_CONF_FILE = os.path.join(DATA_PATH, 'pkgdev.conf')



"""collection of tools for Gentoo development

pkgdev provides a collection of tools for Gentoo development.
"""

from pkgcore.util import commandline

argparser = commandline.ArgumentParser(
    description=__doc__, help=False, subcmds=True, script=(__file__, __name__))



import os
import multiprocessing
import re
import signal
from functools import partial
from unittest.mock import patch

from pkgdev.mangle import Mangler
from pkgdev.scripts.pkgdev_commit import Change
import pytest
from snakeoil.cli.exceptions import UserException


def fake_change(s):
    return Change('/repo', 'A', str(s))


class TestMangler:

    def test_nonexistent_file(self, tmp_path):
        path = tmp_path / 'nonexistent'
        assert list(Mangler([fake_change(path)])) == []

    def test_empty_file(self, tmp_path):
        path = tmp_path / 'empty'
        path.touch()
        assert list(Mangler([fake_change(path)])) == []

    def test_skipped_file(self, tmp_path):
        paths = [(tmp_path / x) for x in ('file', 'file.patch')]

        for p in paths:
            p.write_text('# comment')
        # skip patch files
        skip_regex = re.compile(r'.+\.patch$')
        mangled_paths = set(Mangler(map(fake_change, paths), skip_regex=skip_regex))
        assert mangled_paths == {str(tmp_path / 'file')}

        for p in paths:
            p.write_text('# comment')
        # don't skip any files
        mangled_paths = set(Mangler(map(fake_change, paths)))
        assert mangled_paths == set(map(str, paths))

    def test_nonmangled_file(self, tmp_path):
        path = tmp_path / 'file'
        path.write_text('# comment\n')
        assert list(Mangler([fake_change(path)])) == []

    def test_mangled_file(self, tmp_path):
        path = tmp_path / 'file'
        path.write_text('# comment')
        assert list(Mangler([fake_change(path)])) == [str(path)]
        assert path.read_text() == '# comment\n'

    def test_iterator_exceptions(self, tmp_path):
        """Test parallelized iterator against unhandled exceptions."""
        path = tmp_path / 'file'
        path.write_text('# comment\n')

        def _mangle_func(self, data):
            raise Exception('func failed')

        with patch('pkgdev.mangle.Mangler._mangle', _mangle_func):
            with pytest.raises(UserException, match='Exception: func failed'):
                list(Mangler([fake_change(path)]))

    def test_sigint_handling(self, tmp_path):
        """Verify SIGINT is properly handled by the parallelized pipeline."""
        path = tmp_path / 'file'
        path.write_text('# comment\n')

        def run(queue):
            """Mangler run in a separate process that gets interrupted."""
            import sys
            import time
            from unittest.mock import patch

            from pkgdev.mangle import Mangler

            def sleep():
                """Notify testing process then sleep."""
                queue.put('ready')
                time.sleep(100)

            with patch('pkgdev.mangle.Mangler.__iter__') as fake_iter:
                fake_iter.side_effect = partial(sleep)
                try:
                    iter(Mangler([fake_change(path)]))
                except KeyboardInterrupt:
                    queue.put(None)
                    sys.exit(0)
                queue.put(None)
                sys.exit(1)

        mp_ctx = multiprocessing.get_context('fork')
        queue = mp_ctx.SimpleQueue()
        p = mp_ctx.Process(target=run, args=(queue,))
        p.start()
        # wait for pipeline object to be fully initialized then send SIGINT
        for _ in iter(queue.get, None):
            os.kill(p.pid, signal.SIGINT)
            p.join()
            assert p.exitcode == 0



import os
import shutil
import textwrap
from datetime import datetime
from functools import partial
from io import StringIO
from unittest.mock import patch

import pytest
from pkgdev.mangle import copyright_regex, keywords_regex
from pkgdev.scripts import run
from snakeoil.contexts import chdir, os_environ
from snakeoil.osutils import pjoin


class TestPkgdevCommitParseArgs:

    def test_non_repo_cwd(self, capsys, tool):
        with pytest.raises(SystemExit) as excinfo:
            tool.parse_args(['commit'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev commit: error: not in ebuild repo'

    def test_bad_repo_cwd(self, make_repo, capsys, tool):
        repo = make_repo(masters=('nonexistent',))
        with pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            tool.parse_args(['commit'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.strip().startswith('pkgdev commit: error: repo init failed')

    def test_non_git_repo_cwd(self, repo, capsys, tool):
        with pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            tool.parse_args(['commit'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev commit: error: not in git repo'

    def test_non_ebuild_git_repo_cwd(self, make_repo, git_repo, capsys, tool):
        os.mkdir(pjoin(git_repo.path, 'repo'))
        repo = make_repo(pjoin(git_repo.path, 'repo'))
        with pytest.raises(SystemExit) as excinfo, \
                chdir(repo.location):
            tool.parse_args(['commit'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert err.strip() == 'pkgdev commit: error: not in ebuild git repo'

    def test_commit_signing(self, repo, make_git_repo, tool):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0', commit=False)
        # signed commits aren't enabled by default
        with chdir(repo.location):
            options, _ = tool.parse_args(['commit', '-u'])
            assert '--signoff' not in options.commit_args
            assert '--gpg-sign' not in options.commit_args

            options, _ = tool.parse_args(['commit', '-u', '--signoff'])
            assert '--signoff' in options.commit_args
            assert '--gpg-sign' not in options.commit_args
        # signed commits enabled by layout.conf setting
        with open(pjoin(git_repo.path, 'metadata/layout.conf'), 'a+') as f:
            f.write('sign-commits = true\n')
        with chdir(repo.location):
            options, _ = tool.parse_args(['commit', '-u'])
            assert '--signoff' not in options.commit_args
            assert '--gpg-sign' in options.commit_args

            options, _ = tool.parse_args(['commit', '-u', '--signoff'])
            assert '--signoff' in options.commit_args
            assert '--gpg-sign' in options.commit_args

    def test_git_commit_args(self, repo, make_git_repo, tool):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0', commit=False)
        with chdir(repo.location):
            for opt, expected in (
                    ('-n', '--dry-run'),
                    ('--dry-run', '--dry-run'),
                    ('-v', '-v'),
                    ('--verbose', '-v'),
                    ):
                options, _ = tool.parse_args(['commit', opt])
                assert expected in options.commit_args

    def test_git_commit_args_passthrough(self, repo, make_git_repo, tool):
        """Unknown arguments for ``pkgdev commit`` are passed to ``git commit``."""
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0', commit=False)
        with chdir(repo.location):
            for opt in ('--author="A U Thor <author@example.com>"', '-e'):
                options, _ = tool.parse_args(['commit', opt])
            assert options.commit_args == [opt]

    def test_scan_args(self, repo, make_git_repo, tool):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0', commit=False)
        # pkgcheck isn't run in verbose mode by default
        with chdir(repo.location):
            options, _ = tool.parse_args(['commit'])
        assert '-v' not in options.scan_args
        # verbosity level is passed down to pkgcheck
        with chdir(repo.location):
            options, _ = tool.parse_args(['commit', '-v'])
        assert '-v' in options.scan_args

    def test_commit_tags(self, capsys, repo, make_git_repo, tool):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0', commit=False)
        with chdir(repo.location):
            # bug IDs
            for opt in ('-b', '--bug'):
                options, _ = tool.parse_args(['commit', opt, '1'])
                assert options.footer == {('Bug', 'https://bugs.gentoo.org/1')}

            # bug URLs
            for opt in ('-b', '--bug'):
                options, _ = tool.parse_args(['commit', opt, 'https://bugs.gentoo.org/2'])
                assert options.footer == {('Bug', 'https://bugs.gentoo.org/2')}

            # bug IDs
            for opt in ('-c', '--closes'):
                options, _ = tool.parse_args(['commit', opt, '1'])
                assert options.footer == {('Closes', 'https://bugs.gentoo.org/1')}

            # bug URLs
            for opt in ('-c', '--closes'):
                options, _ = tool.parse_args(['commit', opt, 'https://bugs.gentoo.org/2'])
                assert options.footer == {('Closes', 'https://bugs.gentoo.org/2')}

            # bad URL
            for opt in ('-b', '-c'):
                with pytest.raises(SystemExit) as excinfo:
                    tool.parse_args(['commit', opt, 'bugs.gentoo.org/1'])
                assert excinfo.value.code == 2
                out, err = capsys.readouterr()
                assert not out
                assert 'invalid URL: bugs.gentoo.org/1' in err

            # generic tags
            for opt in ('-T', '--tag'):
                for value, expected in (
                        ('tag:value', ('Tag', 'value')),
                        ('tag:multiple values', ('Tag', 'multiple values')),
                        ('tag:multiple:values', ('Tag', 'multiple:values')),
                        ):
                    options, _ = tool.parse_args(['commit', opt, value])
                    assert options.footer == {expected}

            # bad tags
            for opt in ('-T', '--tag'):
                for value in ('', ':', 'tag:', ':value', 'tag'):
                    with pytest.raises(SystemExit) as excinfo:
                        tool.parse_args(['commit', opt, value])
                    assert excinfo.value.code == 2
                    out, err = capsys.readouterr()
                    assert not out
                    assert 'invalid commit tag' in err


class TestPkgdevCommit:

    script = staticmethod(partial(run, 'pkgdev'))

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path):
        self.cache_dir = str(tmp_path)
        self.scan_args = ['--pkgcheck-scan', f'--config no --cache-dir {self.cache_dir}']
        # args for running pkgdev like a script
        self.args = ['pkgdev', 'commit', '--config', 'no'] + self.scan_args

    def test_empty_repo(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location, commit=True)
        with patch('sys.argv', self.args), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert not out
        assert err.strip() == 'pkgdev commit: error: no staged changes exist'

    def test_git_message_opts(self, repo, make_git_repo, tmp_path):
        """Verify message-related options are passed through to `git commit`."""
        git_repo = make_git_repo(repo.location, commit=True)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0', commit=False)
        path = str(tmp_path / 'msg')
        with open(path, 'w') as f:
            f.write('commit1')

        with patch('sys.argv', self.args + ['-u', '-F', path]), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 0
        commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
        assert commit_msg == ['commit1']

        repo.create_ebuild('cat/pkg-1')
        git_repo.add_all('cat/pkg-1', commit=False)
        with os_environ(GIT_EDITOR="sed -i '1s/1/2/'"), \
                patch('sys.argv', self.args + ['-u', '-t', path]), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 0
        commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
        assert commit_msg == ['commit2']

    def test_message_template(self, capsys, repo, make_git_repo, tmp_path):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        path = str(tmp_path / 'msg')

        # auto-generate prefix
        with open(path, 'w') as f:
            f.write(textwrap.dedent("""\
                *: summary

                body
            """))

        for i, opt in enumerate(['-M', '--message-template'], 1):
            repo.create_ebuild(f'cat/pkg-{i}')
            git_repo.add_all(f'cat/pkg-{i}', commit=False)
            with patch('sys.argv', self.args + ['-u', opt, path]), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
            assert commit_msg == ['cat/pkg: summary', '', 'body']

        # override prefix
        with open(path, 'w') as f:
            f.write(textwrap.dedent("""\
                prefix: summary

                body
            """))

        for i, opt in enumerate(['-M', '--message-template'], 3):
            repo.create_ebuild(f'cat/pkg-{i}')
            git_repo.add_all(f'cat/pkg-{i}', commit=False)
            with patch('sys.argv', self.args + ['-u', opt, path]), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
            assert commit_msg == ['prefix: summary', '', 'body']

        # empty message
        with open(path, 'w') as f:
            f.write('')

        for i, opt in enumerate(['-M', '--message-template'], 5):
            repo.create_ebuild(f'cat/pkg-{i}')
            git_repo.add_all(f'cat/pkg-{i}', commit=False)
            with patch('sys.argv', self.args + ['-u', opt, path]), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 2
            out, err = capsys.readouterr()
            assert not out
            assert err.strip().startswith('pkgdev commit: error: empty message template')

    def test_custom_unprefixed_message(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        ebuild_path = repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        with open(ebuild_path, 'a+') as f:
            f.write('# comment\n')

        with patch('sys.argv', self.args + ['-u', '-m', 'msg']), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 0
        out, err = capsys.readouterr()
        assert err == out == ''

        commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
        assert commit_msg == ['cat/pkg: msg']

    def test_custom_prefixed_message(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        ebuild_path = repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        with open(ebuild_path, 'a+') as f:
            f.write('# comment\n')

        with patch('sys.argv', self.args + ['-u', '-m', 'prefix: msg']), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 0
        out, err = capsys.readouterr()
        assert err == out == ''

        commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
        assert commit_msg == ['prefix: msg']

    def test_edited_commit_message(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        ebuild_path = repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        with open(ebuild_path, 'a+') as f:
            f.write('# comment\n')

        with os_environ(GIT_EDITOR="sed -i '1s/$/commit/'"), \
                patch('sys.argv', self.args + ['-u']), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 0
        out, err = capsys.readouterr()
        assert err == out == ''

        commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
        assert commit_msg == ['cat/pkg: commit']

    def test_generated_commit_prefixes(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')

        def commit():
            with patch('sys.argv', self.args + ['-a', '-m', 'msg']), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            out, err = capsys.readouterr()
            assert err == out == ''
            message = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
            return message[0]

        # single package change
        repo.create_ebuild('cat/newpkg-0')
        assert commit().startswith('cat/newpkg: ')

        # multiple package changes in the same category
        repo.create_ebuild('cat/newpkg-1')
        repo.create_ebuild('cat/pkg-1')
        assert commit().startswith('cat/*: ')

        # multiple package changes in various categories
        repo.create_ebuild('cat/newpkg-2')
        repo.create_ebuild('cat/pkg-2')
        repo.create_ebuild('newcat/newpkg-1')
        assert commit().startswith('*/*: ')

        # single eclass change
        with open(pjoin(repo.location, 'eclass', 'foo.eclass'), 'a+') as f:
            f.write('# comment\n')
        assert commit().startswith('foo.eclass: ')

        # multiple eclass changes
        for eclass in ('foo.eclass', 'bar.eclass'):
            with open(pjoin(repo.location, 'eclass', eclass), 'a+') as f:
                f.write('# comment\n')
        assert commit().startswith('eclass: ')

        # single profiles/package.mask change
        with open(pjoin(repo.location, 'profiles', 'package.mask'), 'a+') as f:
            f.write('# comment\n')
        assert commit().startswith('profiles: ')

        amd64 = pjoin(repo.location, 'profiles', 'arch', 'amd64')
        os.makedirs(amd64)
        arm64 = pjoin(repo.location, 'profiles', 'arch', 'arm64')
        os.makedirs(arm64)

        # multiple profiles file changes in the same subdir
        for file in ('package.mask', 'package.mask'):
            with open(pjoin(amd64, file), 'a+') as f:
                f.write('# comment\n')
        assert commit().startswith('profiles/arch/amd64: ')

        # multiple profiles file changes in different subdirs
        for path in (amd64, arm64):
            with open(pjoin(path, 'package.use'), 'a+') as f:
                f.write('# comment\n')
        assert commit().startswith('profiles/arch: ')

        # single repo root file change
        with open(pjoin(repo.location, 'skel.ebuild'), 'a+') as f:
            f.write('# comment\n')
        assert commit().startswith('skel.ebuild: ')

        # multiple repo root file change (no commit message prefix)
        for file in ('skel.ebuild', 'header.txt'):
            with open(pjoin(repo.location, file), 'a+') as f:
                f.write('# comment\n')
        assert commit() == 'msg'

        # treewide changes (no commit message prefix)
        repo.create_ebuild('foo/bar-1')
        with open(pjoin(repo.location, 'eclass', 'foo.eclass'), 'a+') as f:
            f.write('# comment\n')
        with open(pjoin(repo.location, 'profiles', 'package.mask'), 'a+') as f:
            f.write('# comment\n')
        assert commit() == 'msg'

    def test_generated_commit_summaries(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')

        def commit():
            with os_environ(GIT_EDITOR="sed -i '1s/$/summary/'"), \
                    patch('sys.argv', self.args + ['-a']), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            out, err = capsys.readouterr()
            assert err == out == ''
            message = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
            return message[0]

        # initial package import
        repo.create_ebuild('cat/newpkg-0')
        assert commit() == 'cat/newpkg: new package, add 0'

        # initial package import, overflowed title truncated
        for i in range(10):
            repo.create_ebuild(f'cat/newpkg2-{i}.0.0')
        assert commit() == 'cat/newpkg2: new package'

        # single addition
        repo.create_ebuild('cat/pkg-1')
        assert commit() == 'cat/pkg: add 1'

        # multiple additions
        repo.create_ebuild('cat/pkg-2')
        repo.create_ebuild('cat/pkg-3')
        repo.create_ebuild('cat/pkg-4', eapi=6)
        assert commit() == 'cat/pkg: add 2, 3, 4'

        # revbump updating EAPI
        repo.create_ebuild('cat/pkg-4-r1', eapi=7)
        assert commit() == 'cat/pkg: update EAPI 6 -> 7'

        # single rename with no revisions
        git_repo.move(
            pjoin(git_repo.path, 'cat/pkg/pkg-4.ebuild'),
            pjoin(git_repo.path, 'cat/pkg/pkg-5.ebuild'),
            commit=False
        )
        assert commit() == 'cat/pkg: add 5, drop 4'

        # bump EAPI
        repo.create_ebuild('cat/pkg-6', eapi='6')
        git_repo.add_all('cat/pkg-6')
        repo.create_ebuild('cat/pkg-6', eapi='7')
        assert commit() == 'cat/pkg: update EAPI 6 -> 7'

        # update description
        repo.create_ebuild('cat/pkg-7')
        git_repo.add_all('cat/pkg-7')
        repo.create_ebuild('cat/pkg-7', description='something')
        assert commit() == 'cat/pkg: update DESCRIPTION'

        # update description & homepage
        repo.create_ebuild('cat/pkg-7', description='another something', homepage='https://gentoo.org')
        assert commit() == 'cat/pkg: update DESCRIPTION, HOMEPAGE'

        # update string_targets (USE_RUBY)
        repo.create_ebuild('cat/pkg-8', use_ruby='ruby27')
        git_repo.add_all('cat/pkg-8')
        repo.create_ebuild('cat/pkg-8', use_ruby='ruby27 ruby30')
        assert commit() == 'cat/pkg: enable ruby30'
        repo.create_ebuild('cat/pkg-8', use_ruby='ruby30')
        assert commit() == 'cat/pkg: disable ruby27'
        repo.create_ebuild('cat/pkg-8', use_ruby=' '.join(f'ruby{i}' for i in range(30, 40)))
        assert commit() == 'cat/pkg: update USE_RUBY support'

        # update array_targets (PYTHON_COMPAT)
        repo.create_ebuild('cat/pkg-9', data='PYTHON_COMPAT=( python3_9 )')
        git_repo.add_all('cat/pkg-9')
        repo.create_ebuild('cat/pkg-9', data='PYTHON_COMPAT=( python3_{9..10} )')
        assert commit() == 'cat/pkg: enable py3.10'
        repo.create_ebuild('cat/pkg-9', data='PYTHON_COMPAT=( python3_10 )')
        assert commit() == 'cat/pkg: disable py3.9'


        # multiple ebuild modifications don't get a generated summary
        repo.create_ebuild('cat/pkg-5', keywords=['~amd64'])
        repo.create_ebuild('cat/pkg-6', keywords=['~amd64'])
        assert commit() == 'cat/pkg: summary'

        # large number of additions in a single commit
        for v in range(10000, 10010):
            repo.create_ebuild(f'cat/pkg-{v}')
        assert commit() == 'cat/pkg: add versions'

        # create Manifest
        with open(pjoin(git_repo.path, 'cat/pkg/Manifest'), 'w') as f:
            f.write('DIST pkg-3.tar.gz 101 BLAKE2B deadbeef SHA512 deadbeef\n')
        assert commit() == 'cat/pkg: update Manifest'
        # update Manifest
        with open(pjoin(git_repo.path, 'cat/pkg/Manifest'), 'a+') as f:
            f.write('DIST pkg-2.tar.gz 101 BLAKE2B deadbeef SHA512 deadbeef\n')
        assert commit() == 'cat/pkg: update Manifest'
        # remove Manifest
        os.remove(pjoin(git_repo.path, 'cat/pkg/Manifest'))
        assert commit() == 'cat/pkg: update Manifest'

        # single removals
        os.remove(pjoin(git_repo.path, 'cat/pkg/pkg-4-r1.ebuild'))
        assert commit() == 'cat/pkg: drop 4-r1'
        os.remove(pjoin(git_repo.path, 'cat/pkg/pkg-3.ebuild'))
        assert commit() == 'cat/pkg: drop 3'

        # multiple removal
        os.remove(pjoin(git_repo.path, 'cat/pkg/pkg-2.ebuild'))
        os.remove(pjoin(git_repo.path, 'cat/pkg/pkg-1.ebuild'))
        assert commit() == 'cat/pkg: drop 1, 2'

        # large number of removals in a single commit
        for v in range(10000, 10010):
            os.remove(pjoin(git_repo.path, f'cat/pkg/pkg-{v}.ebuild'))
        assert commit() == 'cat/pkg: drop versions'

        # package rename
        shutil.copytree(pjoin(git_repo.path, 'cat/pkg'), pjoin(git_repo.path, 'newcat/pkg'))
        shutil.rmtree(pjoin(git_repo.path, 'cat/pkg'))
        assert commit() == 'newcat/pkg: rename cat/pkg'

        # treeclean
        shutil.rmtree(pjoin(git_repo.path, 'newcat/pkg'))
        assert commit() == 'newcat/pkg: treeclean'

    def test_generated_commit_summaries_keywords(self, capsys, make_repo, make_git_repo):
        repo = make_repo(arches=['amd64', 'arm64', 'ia64', 'x86'])
        git_repo = make_git_repo(repo.location)
        pkgdir = os.path.dirname(repo.create_ebuild('cat/pkg-0'))
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <stabilize-allarches/>
                </pkgmetadata>
            """))
        git_repo.add_all('cat/pkg-0')

        def commit():
            with os_environ(GIT_EDITOR="sed -i '1s/$/summary/'"), \
                    patch('sys.argv', self.args + ['-a']), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            out, err = capsys.readouterr()
            assert err == out == ''
            message = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
            return message[0]

        # keyword version
        repo.create_ebuild('cat/pkg-0', keywords=['~amd64'])
        assert commit() == 'cat/pkg: keyword 0 for ~amd64'

        # stabilize version
        repo.create_ebuild('cat/pkg-0', keywords=['amd64'])
        assert commit() == 'cat/pkg: stabilize 0 for amd64'

        # destabilize version
        repo.create_ebuild('cat/pkg-0', keywords=['~amd64'])
        assert commit() == 'cat/pkg: destabilize 0 for ~amd64'

        # unkeyword version
        repo.create_ebuild('cat/pkg-0')
        assert commit() == 'cat/pkg: unkeyword 0 for ~amd64'

        with open(pjoin(repo.location, 'profiles', 'arches.desc'), 'w') as f:
            f.write(textwrap.dedent("""\
                amd64 stable
                arm64 stable
                ia64  testing
                x86   stable
            """))
        git_repo.add_all('set arches.desc')

        repo.create_ebuild('cat/pkg-0', keywords=['~amd64', '~arm64', '~ia64', '~x86'])
        git_repo.add_all('cat/pkg-0')

        # stabilize version
        repo.create_ebuild('cat/pkg-0', keywords=['amd64', '~arm64', '~ia64', '~x86'])
        assert commit() == 'cat/pkg: stabilize 0 for amd64'

        # stabilize version ALLARCHES
        repo.create_ebuild('cat/pkg-0', keywords=['amd64', 'arm64', '~ia64', 'x86'])
        assert commit() == 'cat/pkg: stabilize 0 for ALLARCHES'

        # stabilize version
        repo.create_ebuild('cat/newpkg-0', keywords=['~amd64', '~arm64', '~ia64', '~x86'])
        git_repo.add_all('cat/newpkg')
        repo.create_ebuild('cat/newpkg-0', keywords=['amd64', 'arm64', '~ia64', 'x86'])
        assert commit() == 'cat/newpkg: stabilize 0 for amd64, arm64, x86'

    def test_metadata_summaries(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        pkgdir = os.path.dirname(repo.create_ebuild('cat/pkg-0'))
        # stub metadata
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                </pkgmetadata>
            """))
        git_repo.add_all('cat/pkg-0')

        def commit():
            with os_environ(GIT_EDITOR="sed -i '1s/$/summary/'"), \
                    patch('sys.argv', self.args + ['-a']), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            out, err = capsys.readouterr()
            assert err == out == ''
            message = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
            return message[0]

        # add yourself
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                    <maintainer type="person">
                        <email>first.last@email.com</email>
                        <name>First Last</name>
                    </maintainer>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: add myself as a maintainer'

        # drop yourself
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: drop myself as a maintainer'

        # drop to maintainer-needed
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: drop to maintainer-needed'

        # add random maintainer
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: update maintainers'

        # add allarches tag
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                    <stabilize-allarches/>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: mark ALLARCHES'

        # drop allarches tag
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: drop ALLARCHES'

        # add upstream metadata
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                    <upstream>
                        <remote-id type="github">pkgcore/pkgdev</remote-id>
                        <remote-id type="pypi">pkgdev</remote-id>
                    </upstream>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: add github, pypi upstream metadata'

        # remove upstream metadata
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                    <upstream>
                        <remote-id type="github">pkgcore/pkgdev</remote-id>
                    </upstream>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: remove pypi upstream metadata'

        # update upstream metadata
        with open(pjoin(pkgdir, 'metadata.xml'), 'w') as f:
            f.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "http://www.gentoo.org/dtd/metadata.dtd">
                <pkgmetadata>
                    <maintainer type="person">
                        <email>person@email.com</email>
                        <name>Person</name>
                    </maintainer>
                    <upstream>
                        <remote-id type="github">pkgcore/pkgcheck</remote-id>
                    </upstream>
                </pkgmetadata>
            """))
        assert commit() == 'cat/pkg: update upstream metadata'

    def test_no_summary(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')

        def commit(args):
            with os_environ(GIT_EDITOR="sed -i '1s/$/summary/'"), \
                    patch('sys.argv', self.args + args), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            out, err = capsys.readouterr()
            assert err == out == ''
            return git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])

        # no commit message content
        for i in range(10):
            with open(pjoin(git_repo.path, f'file-a-{i}'), 'w') as f:
                f.write('stub\n')
        assert commit(['-a']) == ['summary']

        # footer exists with no generated summary
        for i in range(10):
            with open(pjoin(git_repo.path, f'file-b-{i}'), 'w') as f:
                f.write('stub\n')
        assert commit(['-a', '-T', 'tag:value']) == ['summary', '', 'Tag: value']

    def test_non_gentoo_file_mangling(self, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        ebuild_path = repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')

        def commit(args):
            with patch('sys.argv', self.args + args), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0

        # append line missing EOF newline to ebuild
        with open(ebuild_path, 'a+') as f:
            f.write('# comment')
        # verify file doesn't end with newline
        with open(ebuild_path) as f:
            assert f.read()[-1] != '\n'

        # non-gentoo repos aren't mangled by default
        commit(['-u', '-m', 'mangling'])
        with open(ebuild_path) as f:
            assert f.read()[-1] != '\n'

        # but they can be forcibly mangled
        with open(ebuild_path, 'a+') as f:
            f.write('# comment')
        commit(['--mangle', '-u', '-m', 'mangling'])
        # mangled pre-commit, file now ends with newline
        with open(ebuild_path) as f:
            assert f.read()[-1] == '\n'

        # FILESDIR content is ignored even when forced
        path = pjoin(os.path.dirname(ebuild_path), 'files', 'pkg.patch')
        os.makedirs(os.path.dirname(path))
        with open(path, 'w') as f:
            f.write('# comment')
        # verify file doesn't end with newline
        with open(path) as f:
            assert f.read()[-1] != '\n'

    def test_gentoo_file_mangling(self, make_repo, make_git_repo):
        repo = make_repo(repo_id='gentoo')
        git_repo = make_git_repo(repo.location)
        ebuild_path = repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')

        def commit(args):
            with patch('sys.argv', self.args + args), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0

        # append line missing EOF newline to ebuild
        with open(ebuild_path, 'a+') as f:
            f.write('# comment')
        # verify file doesn't end with newline
        with open(ebuild_path) as f:
            assert f.read()[-1] != '\n'

        # gentoo repos are mangled by default
        commit(['-n', '-u', '-m', 'mangling'])
        with open(ebuild_path) as f:
            assert f.read()[-1] == '\n'

        # FILESDIR content is ignored
        path = pjoin(os.path.dirname(ebuild_path), 'files', 'pkg.patch')
        os.makedirs(os.path.dirname(path))
        with open(path, 'w') as f:
            f.write('# comment')
        # verify file doesn't end with newline
        with open(path) as f:
            assert f.read()[-1] != '\n'

        for years, org in (
                ('1999-2020', 'Gentoo Authors'),
                ('1999-2020', 'Gentoo Foundation'),
                ('2020', 'Gentoo Authors'),
                ('2020', 'Gentoo Foundation'),
                ):
            # munge the copyright header
            with open(ebuild_path, 'r+') as f:
                lines = f.read().splitlines()
                lines[0] = f'# Copyright {years} {org}\n'
                f.seek(0)
                f.truncate()
                f.write('\n'.join(lines) + '\n')
            commit(['-n', '-u', '-m', 'mangling'])
            # verify the copyright header was updated
            with open(ebuild_path) as f:
                lines = f.read().splitlines()
                mo = copyright_regex.match(lines[0])
                assert mo.group('end') == str(datetime.today().year)
                assert mo.group('begin') == years[:4] + '-'
                assert mo.group('holder') == 'Gentoo Authors'

        for original, expected in (
                ('"arm64 amd64 x86"', 'amd64 arm64 x86'),
                ('"arm64 amd64 ~x86"', 'amd64 arm64 ~x86'),
                ('"arm64 ~x86 amd64"', 'amd64 arm64 ~x86'),
                ('"arm64 ~x86 ~amd64"', '~amd64 arm64 ~x86'),
                ('arm64 ~x86 ~amd64', '~amd64 arm64 ~x86'),
                ):
            # munge the keywords
            with open(ebuild_path, 'r+') as f:
                lines = f.read().splitlines()
                lines[-1] = f'KEYWORDS={original}'
                f.seek(0)
                f.truncate()
                f.write('\n'.join(lines) + '\n')
            commit(['-n', '-u', '-m', 'mangling'])
            # verify the keywords were updated
            with open(ebuild_path) as f:
                lines = f.read().splitlines()
                mo = keywords_regex.match(lines[-1])
                assert mo.group('keywords') == expected

    def test_scan(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')

        for i, opt in enumerate(['-s', '--scan'], 1):
            repo.create_ebuild(f'cat/pkg-{i}')
            git_repo.add_all(f'cat/pkg-{i}', commit=False)
            with patch('sys.argv', self.args + [opt]), \
                    pytest.raises(SystemExit) as excinfo, \
                    chdir(git_repo.path):
                self.script()
            assert excinfo.value.code == 0
            out, err = capsys.readouterr()
            assert err == out == ''
            commit_msg = git_repo.log(['-1', '--pretty=tformat:%B', 'HEAD'])
            assert commit_msg == [f'cat/pkg: add {i}']

    def test_failed_scan(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')

        # verify staged changes via `pkgcheck scan` before creating commit
        repo.create_ebuild('cat/pkg-1', license='')
        git_repo.add_all('cat/pkg-1', commit=False)
        with patch('sys.argv', self.args + ['--scan']), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 1
        out, err = capsys.readouterr()
        assert not err
        assert out == textwrap.dedent("""\
            cat/pkg
              MissingLicense: version 1: no license defined

            FAILURES
            cat/pkg
              MissingLicense: version 1: no license defined
        """)

        # ignore failures to create the commit
        with patch('sys.argv', self.args + ['--scan', '--ask']), \
                patch('sys.stdin', StringIO('y\n')), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 0

    def test_config_opts(self, capsys, repo, make_git_repo, tmp_path):
        config_file = str(tmp_path / 'config')
        with open(config_file, 'w') as f:
            f.write(textwrap.dedent("""
                [DEFAULT]
                commit.scan=
            """))

        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        repo.create_ebuild('cat/pkg-1', license='')
        git_repo.add_all('cat/pkg-1', commit=False)
        with patch('sys.argv', ['pkgdev', 'commit', '--config', config_file] + self.scan_args), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        out, err = capsys.readouterr()
        assert excinfo.value.code == 1
        assert not err
        assert 'MissingLicense' in out

    def test_config_repo_opts(self, capsys, repo, make_git_repo, tmp_path):
        config_file = str(tmp_path / 'config')
        with open(config_file, 'w') as f:
            f.write(textwrap.dedent("""
                [fake]
                commit.scan=
            """))

        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        repo.create_ebuild('cat/pkg-1', license='')
        git_repo.add_all('cat/pkg-1', commit=False)
        with patch('sys.argv', ['pkgdev', 'commit', '--config', config_file] + self.scan_args), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        out, err = capsys.readouterr()
        assert excinfo.value.code == 1
        assert not err
        assert 'MissingLicense' in out

    def test_failed_manifest(self, capsys, repo, make_git_repo):
        git_repo = make_git_repo(repo.location)
        repo.create_ebuild('cat/pkg-0')
        git_repo.add_all('cat/pkg-0')
        repo.create_ebuild('cat/pkg-1', eapi='-1')
        git_repo.add_all('cat/pkg-1', commit=False)
        with patch('sys.argv', self.args), \
                pytest.raises(SystemExit) as excinfo, \
                chdir(git_repo.path):
            self.script()
        assert excinfo.value.code == 1
        out, err = capsys.readouterr()
        assert not err
        assert out == " * cat/pkg-1: invalid EAPI '-1'\n"



import subprocess
import sys

from snakeoil.cli.exceptions import UserException


class GitError(SystemExit):
    """Generic error running a git command."""


def run(*args, **kwargs):
    """Wrapper for running git via subprocess.run()."""
    kwargs.setdefault('check', True)
    kwargs.setdefault('text', True)
    cmd = ['git'] + list(args)

    # output git command that would be run to stderr
    if '--dry-run' in args:
        git_cmd = ' '.join(x for x in cmd if x != '--dry-run')
        sys.stderr.write(f'{git_cmd}\n')

    try:
        return subprocess.run(cmd, **kwargs)
    except FileNotFoundError as e:
        raise UserException(str(e))
    except subprocess.CalledProcessError as e:
        raise GitError(e.returncode)



import os
import re
import shlex
import subprocess
import tempfile
import textwrap
from collections import deque
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from itertools import groupby
from operator import itemgetter
from typing import List

from pkgcore.ebuild.atom import MalformedAtom
from pkgcore.ebuild.atom import atom as atom_cls
from pkgcore.ebuild.profiles import ProfileNode
from snakeoil.bash import iter_read_bash
from snakeoil.cli import arghparse
from snakeoil.osutils import pjoin
from snakeoil.strings import pluralism

from .. import git
from .argparsers import cwd_repo_argparser, git_repo_argparser

mask = arghparse.ArgumentParser(
    prog='pkgdev mask', description='mask packages',
    parents=(cwd_repo_argparser, git_repo_argparser))
mask.add_argument(
    'targets', metavar='TARGET', nargs='*',
    help='package to mask',
    docs="""
        Packages matching any of these restrictions will have a mask entry in
        profiles/package.mask added for them. If no target is specified a path
        restriction is created based on the current working directory. In other
        words, if ``pkgdev mask`` is run within an ebuild's directory, all the
        ebuilds within that directory will be masked.
    """)
mask_opts = mask.add_argument_group('mask options')
mask_opts.add_argument(
    '-r', '--rites', metavar='DAYS', nargs='?', const=30, type=arghparse.positive_int,
    help='mark for last rites',
    docs="""
        Mark a mask entry for last rites. This defaults to 30 days until
        package removal but accepts an optional argument for the number of
        days.
    """)
mask_opts.add_argument(
    '-b', '--bug', dest='bugs', action='append', type=arghparse.positive_int,
    help='reference bug in the mask comment',
    docs="""
        Add a reference to a bug in the mask comment. May be specified multiple
        times to reference multiple bugs.
    """)
mask_opts.add_argument(
    '--email', action='store_true',
    help='spawn email composer with prepared email for sending to mailing lists',
    docs="""
        Spawn user's preferred email composer with a prepared email for
        sending a last rites message to Gentoo's mailing list (``gentoo-dev``
        and ``gentoo-dev-announce``). The user should manually set the Reply-to
        field for the message to be accepted by ``gentoo-dev-announce``.

        For spawning the preferred email composer, the ``xdg-email`` tool from
        ``x11-misc/xdg-utils`` package.
    """)


@mask.bind_final_check
def _mask_validate(parser, namespace):
    atoms = []

    if namespace.email and not namespace.rites:
        mask.error('last rites required for email support')

    if namespace.targets:
        for x in namespace.targets:
            if os.path.exists(x) and x.endswith('.ebuild'):
                restrict = namespace.repo.path_restrict(x)
                pkg = next(namespace.repo.itermatch(restrict))
                atom = pkg.versioned_atom
            else:
                try:
                    atom = atom_cls(x)
                except MalformedAtom:
                    mask.error(f'invalid atom: {x!r}')
                if not namespace.repo.match(atom):
                    mask.error(f'no repo matches: {x!r}')
            atoms.append(atom)
    else:
        restrict = namespace.repo.path_restrict(os.getcwd())
        # repo, category, and package level restricts
        if len(restrict) != 3:
            mask.error('not in a package directory')
        pkg = next(namespace.repo.itermatch(restrict))
        atoms.append(pkg.unversioned_atom)

    namespace.atoms = sorted(atoms)


@dataclass(frozen=True)
class Mask:
    """Entry in package.mask file."""
    author: str
    email: str
    date: str
    comment: List[str]
    atoms: List[atom_cls]

    _removal_re = re.compile(r'^Removal: (?P<date>\d{4}-\d{2}-\d{2})')

    def __str__(self):
        lines = [f'# {self.author} <{self.email}> ({self.date})']
        lines.extend(f'# {x}' if x else '#' for x in self.comment)
        lines.extend(map(str, self.atoms))
        return '\n'.join(lines)

    @property
    def removal(self):
        """Pull removal date from comment."""
        if mo := self._removal_re.match(self.comment[-1]):
            return mo.group('date')
        return None


def consecutive_groups(iterable, ordering=lambda x: x):
    """Return an iterable split into separate, consecutive groups."""
    for k, g in groupby(enumerate(iterable), key=lambda x: x[0] - ordering(x[1])):
        yield map(itemgetter(1), g)


class MaskFile:
    """Object representing the contents of a package.mask file."""

    attribution_re = re.compile(r'^(?P<author>.+) <(?P<email>.+)> \((?P<date>\d{4}-\d{2}-\d{2})\)$')

    def __init__(self, path):
        self.path = path
        self.profile = ProfileNode(os.path.dirname(path))
        self.header = []
        self.masks = deque()

        # parse existing mask entries
        try:
            self.parse()
        except FileNotFoundError:
            pass

    def parse(self):
        """Parse the given file into Mask objects."""
        with open(self.path) as f:
            lines = f.readlines()

        # determine mask groups by line number
        mask_map = dict(iter_read_bash(self.path, enum_line=True))
        for mask_lines in map(list, consecutive_groups(mask_map)):
            # use profile's EAPI setting to coerce supported masks
            atoms = [self.profile.eapi_atom(mask_map[x]) for x in mask_lines]

            # pull comment lines above initial mask entry line
            comment = []
            i = mask_lines[0] - 2
            while i >= 0 and (line := lines[i].rstrip()):
                if not line.startswith('# ') and line != '#':
                    mask.error(f'invalid mask entry header, lineno {i + 1}: {line!r}')
                comment.append(line[2:])
                i -= 1
            if not self.header:
                self.header = lines[:i + 1]
            comment = list(reversed(comment))

            # pull attribution data from first comment line
            if mo := self.attribution_re.match(comment[0]):
                author, email, date = mo.group('author'), mo.group('email'), mo.group('date')
            else:
                mask.error(f'invalid author, lineno {i + 2}: {comment[0]!r}')

            self.masks.append(Mask(author, email, date, comment[1:], atoms))

    def add(self, mask):
        """Add a new mask to the file."""
        self.masks.appendleft(mask)

    def write(self):
        """Serialize the registered masks back to the related file."""
        with open(self.path, 'w') as f:
            f.write(f'{self}\n')

    def __str__(self):
        return ''.join(self.header) + '\n\n'.join(map(str, self.masks))


def get_comment(bugs, rites: int):
    """Spawn editor to get mask comment."""
    tmp = tempfile.NamedTemporaryFile(mode='w')
    summary = []
    if rites:
        summary.append(f'Removal: {datetime.now(timezone.utc) + timedelta(days=rites):%Y-%m-%d}.')
    if bugs:
        # Bug(s) #A, #B, #C
        bug_list = ", ".join(f'#{b}' for b in bugs)
        s = pluralism(bugs)
        summary.append(f'Bug{s} {bug_list}.')
    if summary := '  '.join(summary):
        tmp.write(f'\n{summary}')
    tmp.write(textwrap.dedent("""

        # Please enter the mask message. Lines starting with '#' will be ignored.
        #
        # - Best last rites (removal) practices -
        #
        # Include the following info:
        # a) reason for masking
        # b) bug # for the removal (and yes you should have one)
        # c) date of removal (either the date or "in x days")
        #
        # Example:
        #
        # Masked for removal in 30 days.  Doesn't work
        # with new libfoo. Upstream dead, gtk-1, smells
        # funny.
        # Bug #987654
    """))
    tmp.flush()

    editor = shlex.split(os.environ.get('VISUAL', os.environ.get('EDITOR', 'nano')))
    try:
        subprocess.run(editor + [tmp.name], check=True)
    except subprocess.CalledProcessError:
        mask.error('failed writing mask comment')
    except FileNotFoundError:
        mask.error(f'nonexistent editor: {editor[0]!r}')

    with open(tmp.name) as f:
        # strip trailing whitespace from lines
        comment = (x.rstrip() for x in f.readlines())
    # strip comments
    comment = (x for x in comment if not x.startswith('#'))
    # strip leading/trailing newlines
    comment = '\n'.join(comment).strip().splitlines()
    if not comment:
        mask.error('empty mask comment')

    return comment


def send_last_rites_email(m: Mask, subject_prefix: str):
    try:
        atoms = ', '.join(map(str, m.atoms))
        subprocess.run(args=[
            'xdg-email', '--utf8',
            '--cc', 'gentoo-dev@lists.gentoo.org',
            '--subject', f'{subject_prefix}: {atoms}',
            '--body', str(m),
            'gentoo-dev-announce@lists.gentoo.org'
        ], check=True)
    except subprocess.CalledProcessError:
        mask.error('failed opening email composer')


@mask.bind_main_func
def _mask(options, out, err):
    mask_file = MaskFile(pjoin(options.repo.location, 'profiles/package.mask'))
    today = datetime.now(timezone.utc)

    # pull name/email from git config
    p = git.run('config', 'user.name', stdout=subprocess.PIPE)
    author = p.stdout.strip()
    p = git.run('config', 'user.email', stdout=subprocess.PIPE)
    email = p.stdout.strip()

    # initial args for Mask obj
    mask_args = {
        'author': author,
        'email': email,
        'date': today.strftime('%Y-%m-%d'),
        'comment': get_comment(options.bugs, options.rites),
        'atoms': options.atoms,
    }

    m = Mask(**mask_args)
    mask_file.add(m)
    mask_file.write()

    if options.email:
        send_last_rites_email(m, 'Last rites')

    return 0



#!/usr/bin/env python3

"""Wrapper for running commandline scripts."""

import os
import sys
from importlib import import_module


def run(script_name):
    """Run a given script module."""
    # Remove the current working directory to avoid implicit
    # namespace package (PEP 420) imports due to directories
    # matching module names.
    try:
        sys.path.remove(os.getcwd())
    except ValueError:
        pass

    try:
        from pkgdev.cli import Tool
        script_module = '.'.join(
            os.path.realpath(__file__).split(os.path.sep)[-3:-1] +
            [script_name.replace('-', '_')])
        script = import_module(script_module)
    except ImportError as e:
        python_version = '.'.join(map(str, sys.version_info[:3]))
        sys.stderr.write(f'Failed importing: {e}!\n')
        sys.stderr.write(
            'Verify that pkgdev and its deps are properly installed '
            f'and/or PYTHONPATH is set correctly for python {python_version}.\n')
        if '--debug' in sys.argv[1:]:
            raise
        sys.stderr.write('Add --debug to the commandline for a traceback.\n')
        sys.exit(1)

    tool = Tool(script.argparser)
    sys.exit(tool())


if __name__ == '__main__':
    # We're in a git repo or tarball so add the src dir to the system path.
    # Note that this assumes a certain module layout.
    src_dir = os.path.realpath(__file__).rsplit(os.path.sep, 3)[0]
    sys.path.insert(0, src_dir)
    run(os.path.basename(__file__))



import textwrap

import pytest
from pkgdev import cli
from snakeoil.cli import arghparse


class TestConfigFileParser:

    @pytest.fixture(autouse=True)
    def _create_argparser(self, tmp_path):
        self.config_file = str(tmp_path / 'config')
        self.parser = arghparse.ArgumentParser(prog='pkgdev cli_test')
        self.namespace = arghparse.Namespace()
        self.config_parser = cli.ConfigFileParser(self.parser)

    def test_no_configs(self):
        config = self.config_parser.parse_config(())
        assert config.sections() == []
        namespace = self.config_parser.parse_config_options(self.namespace)
        assert vars(namespace) == {}

    def test_ignored_configs(self):
        # nonexistent config files are ignored
        config = self.config_parser.parse_config(('foo', 'bar'))
        assert config.sections() == []

    def test_bad_config_format_no_section(self, capsys):
        with open(self.config_file, 'w') as f:
            f.write('foobar\n')
        with pytest.raises(SystemExit) as excinfo:
            self.config_parser.parse_config((self.config_file,))
        out, err = capsys.readouterr()
        assert not out
        assert 'parsing config file failed: File contains no section headers' in err
        assert self.config_file in err
        assert excinfo.value.code == 2

    def test_bad_config_format(self, capsys):
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent("""
                [DEFAULT]
                foobar
            """))
        with pytest.raises(SystemExit) as excinfo:
            self.config_parser.parse_config((self.config_file,))
        out, err = capsys.readouterr()
        assert not out
        assert 'parsing config file failed: Source contains parsing errors' in err
        assert excinfo.value.code == 2

    def test_nonexistent_config_options(self, capsys):
        """Nonexistent parser arguments don't cause errors."""
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent("""
                [DEFAULT]
                cli_test.foo=bar
            """))
        with pytest.raises(SystemExit) as excinfo:
            self.config_parser.parse_config_options(None, configs=(self.config_file,))
        out, err = capsys.readouterr()
        assert not out
        assert 'failed loading config: unknown arguments: --foo=bar' in err
        assert excinfo.value.code == 2

    def test_config_options_other_prog(self):
        self.parser.add_argument('--foo')
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent("""
                [DEFAULT]
                other.foo=bar
            """))
        namespace = self.parser.parse_args(['--foo', 'foo'])
        assert namespace.foo == 'foo'
        # config args don't override not matching namespace attrs
        namespace = self.config_parser.parse_config_options(namespace, configs=[self.config_file])
        assert namespace.foo == 'foo'

    def test_config_options(self):
        self.parser.add_argument('--foo')
        with open(self.config_file, 'w') as f:
            f.write(textwrap.dedent("""
                [DEFAULT]
                cli_test.foo=bar
            """))
        namespace = self.parser.parse_args(['--foo', 'foo'])
        assert namespace.foo == 'foo'
        # config args override matching namespace attrs
        namespace = self.config_parser.parse_config_options(namespace, configs=[self.config_file])
        assert namespace.foo == 'bar'



"""Various command-line specific support."""

import argparse
import configparser
import logging
import os

from pkgcore.util import commandline
from snakeoil.cli import arghparse
from snakeoil.contexts import patch
from snakeoil.klass import jit_attr_none
from snakeoil.mappings import OrderedSet
from pkgcore.repository import errors as repo_errors
from pkgcore.util.commandline import _mk_domain

from . import const


class Tool(commandline.Tool):

    def main(self):
        # suppress all pkgcore log messages
        logging.getLogger('pkgcore').setLevel(100)
        return super().main()


class ConfigArg(argparse._StoreAction):
    """Store config path string or False when explicitly disabled."""

    def __call__(self, parser, namespace, values, option_string=None):
        if values.lower() in ('false', 'no', 'n'):
            values = False
        setattr(namespace, self.dest, values)


class ConfigParser(configparser.ConfigParser):
    """ConfigParser with case-sensitive keys (default forces lowercase)."""

    def optionxform(self, option):
        return option


class ConfigFileParser:
    """Argument parser that supports loading settings from specified config files."""

    default_configs = (const.SYSTEM_CONF_FILE, const.USER_CONF_FILE)

    def __init__(self, parser: arghparse.ArgumentParser, configs=(), **kwargs):
        super().__init__(**kwargs)
        self.parser = parser
        self.configs = OrderedSet(configs)

    @jit_attr_none
    def config(self):
        return self.parse_config()

    def parse_config(self, configs=()):
        """Parse given config files."""
        configs = configs if configs else self.configs
        config = ConfigParser(default_section=None)
        try:
            for f in configs:
                config.read(f)
        except configparser.ParsingError as e:
            self.parser.error(f'parsing config file failed: {e}')
        return config

    def parse_config_sections(self, namespace, sections):
        """Parse options from a given iterable of config section names."""
        assert self.parser.prog.startswith('pkgdev ')
        module = self.parser.prog.split(' ', 1)[1] + '.'
        with patch('snakeoil.cli.arghparse.ArgumentParser.error', self._config_error):
            for section in (x for x in sections if x in self.config):
                config_args = ((k.split('.', 1)[1], v) for k, v in self.config.items(section) if k.startswith(module))
                config_args = (f'--{k}={v}' if v else f'--{k}' for k, v in config_args)
                namespace, args = self.parser.parse_known_optionals(config_args, namespace)
                if args:
                    self.parser.error(f"unknown arguments: {'  '.join(args)}")
        return namespace

    def parse_config_options(self, namespace, configs=()):
        """Parse options from config if they exist."""
        configs = list(filter(os.path.isfile, configs))
        if not configs:
            return namespace

        self.configs.update(configs)
        # reset jit attr to force reparse
        self._config = None

        # load default options
        namespace = self.parse_config_sections(namespace, ['DEFAULT'])

        return namespace

    def _config_error(self, message, status=2):
        """Stub to replace error method that notes config failure."""
        self.parser.exit(status, f'{self.parser.prog}: failed loading config: {message}\n')

class ArgumentParser(arghparse.ArgumentParser):
    """Parse all known arguments, from command line and config file."""

    def __init__(self, parents=(), **kwargs):
        self.config_argparser = arghparse.ArgumentParser(suppress=True)
        config_options = self.config_argparser.add_argument_group('config options')
        config_options.add_argument(
            '--config', action=ConfigArg, dest='config_file',
            help='use custom pkgdev settings file',
            docs="""
                Load custom pkgdev scan settings from a given file.

                Note that custom user settings override all other system and repo-level
                settings.

                It's also possible to disable all types of settings loading by
                specifying an argument of 'false' or 'no'.
            """)
        _mk_domain(config_options)
        super().__init__(parents=[*parents, self.config_argparser], **kwargs)

    def parse_known_args(self, args=None, namespace=None):
        temp_namespace, _ = self.config_argparser.parse_known_args(args, namespace)
        # parser supporting config file options
        config_parser = ConfigFileParser(self)
        # always load settings from bundled config
        namespace = config_parser.parse_config_options(
            namespace, configs=[const.BUNDLED_CONF_FILE])

        # load default args from system/user configs if config-loading is allowed
        if temp_namespace.config_file is None:
            namespace = config_parser.parse_config_options(
                namespace, configs=ConfigFileParser.default_configs)
        elif temp_namespace.config_file is not False:
            namespace = config_parser.parse_config_options(
                namespace, configs=(namespace.config_file, ))

        try:
            repo = temp_namespace.domain.find_repo(
                os.getcwd(), config=temp_namespace.config, configure=False)
            if repo is not None:
                namespace = config_parser.parse_config_sections(namespace, repo.aliases)
        except (repo_errors.InitializationError, IOError) as exc:
            self.error(str(exc))

        # parse command line args to override config defaults
        return super().parse_known_args(args, namespace)



"""display package keywords"""

import os
from functools import partial

from pkgcore.ebuild import restricts
from pkgcore.util import commandline
from pkgcore.util import packages as pkgutils
from snakeoil.strings import pluralism

from .. import cli
from .._vendor.tabulate import tabulate, tabulate_formats


showkw = cli.ArgumentParser(
    prog='pkgdev showkw', description='show package keywords')
showkw.add_argument(
    'targets', metavar='target', nargs='*',
    action=commandline.StoreTarget,
    help='extended atom matching of packages')

output_opts = showkw.add_argument_group('output options')
output_opts.add_argument(
    '-f', '--format', default='showkw', metavar='FORMAT',
    choices=tabulate_formats,
    help='keywords table format',
    docs=f"""
        Output table using specified tabular format (defaults to compressed,
        custom format).

        Available formats: {', '.join(tabulate_formats)}
    """)
output_opts.add_argument(
    '-c', '--collapse', action='store_true',
    help='show collapsed list of arches')

arch_options = showkw.add_argument_group('arch options')
arch_options.add_argument(
    '-s', '--stable', action='store_true',
    help='show stable arches')
arch_options.add_argument(
    '-u', '--unstable', action='store_true',
    help='show unstable arches')
arch_options.add_argument(
    '-o', '--only-unstable', action='store_true',
    help='show arches that only have unstable keywords')
arch_options.add_argument(
    '-p', '--prefix', action='store_true',
    help='show prefix and non-native arches')
arch_options.add_argument(
    '-a', '--arch', action='csv_negations',
    help='select arches to display')

# TODO: allow multi-repo comma-separated input
target_opts = showkw.add_argument_group('target options')
target_opts.add_argument(
    '-r', '--repo', dest='selected_repo', metavar='REPO', priority=29,
    action=commandline.StoreRepoObject,
    repo_type='all-raw', allow_external_repos=True,
    help='repo to query (defaults to all ebuild repos)')
@showkw.bind_delayed_default(30, 'repos')
def _setup_repos(namespace, attr):
    target_repo = namespace.selected_repo
    all_ebuild_repos = namespace.domain.all_ebuild_repos_raw
    namespace.cwd = os.getcwd()

    # TODO: move this to StoreRepoObject
    if target_repo is None:
        # determine target repo from the target directory
        for repo in all_ebuild_repos.trees:
            if namespace.cwd in repo:
                target_repo = repo
                break
        else:
            # determine if CWD is inside an unconfigured repo
            target_repo = namespace.domain.find_repo(
                namespace.cwd, config=namespace.config)

    # fallback to using all, unfiltered ebuild repos if no target repo can be found
    namespace.repo = target_repo if target_repo is not None else all_ebuild_repos


@showkw.bind_delayed_default(40, 'arches')
def _setup_arches(namespace, attr):
    default_repo = namespace.config.get_default('repo')

    try:
        known_arches = {arch for r in namespace.repo.trees
                        for arch in r.config.known_arches}
    except AttributeError:
        try:
            # binary/vdb repos use known arches from the default repo
            known_arches = default_repo.config.known_arches
        except AttributeError:
            # TODO: remove fallback for tests after fixing default repo pull
            # from faked config
            known_arches = set()

    arches = known_arches
    if namespace.arch is not None:
        disabled_arches, enabled_arches = namespace.arch
        disabled_arches = set(disabled_arches)
        enabled_arches = set(enabled_arches)
        unknown_arches = disabled_arches.difference(known_arches) | enabled_arches.difference(known_arches)
        if unknown_arches:
            unknown = ', '.join(map(repr, sorted(unknown_arches)))
            known = ', '.join(sorted(known_arches))
            es = pluralism(unknown_arches, plural='es')
            showkw.error(f'unknown arch{es}: {unknown} (choices: {known})')
        if enabled_arches:
            arches = arches.intersection(enabled_arches)
        if disabled_arches:
            arches = arches - disabled_arches

    prefix_arches = set(x for x in arches if '-' in x)
    native_arches = arches.difference(prefix_arches)
    arches = native_arches
    if namespace.prefix:
        arches = arches.union(prefix_arches)
    if namespace.stable:
        try:
            stable_arches = {arch for r in namespace.repo.trees
                             for arch in r.config.profiles.arches('stable')}
        except AttributeError:
            # binary/vdb repos use stable arches from the default repo
            stable_arches = default_repo.config.profiles.arches('stable')
        arches = arches.intersection(stable_arches)

    namespace.known_arches = known_arches
    namespace.prefix_arches = prefix_arches
    namespace.native_arches = native_arches
    namespace.arches = arches


def _colormap(colors, line):
    if colors is None:
        return line
    return colors[line] + line + colors['reset']


@showkw.bind_final_check
def _validate_args(parser, namespace):
    namespace.pkg_dir = False

    # disable colors when not using the native output format
    if namespace.format != 'showkw':
        namespace.color = False

    if namespace.color:
        # default colors to use for keyword types
        _COLORS = {
            '+': '\u001b[32m',
            '~': '\u001b[33m',
            '-': '\u001b[31m',
            '*': '\u001b[31m',
            'o': '\u001b[30;1m',
            'reset': '\u001b[0m',
        }
    else:
        _COLORS = None
    namespace.colormap = partial(_colormap, _COLORS)

    if not namespace.targets:
        if namespace.selected_repo:
            # use repo restriction since no targets specified
            restriction = restricts.RepositoryDep(namespace.selected_repo.repo_id)
            token = namespace.selected_repo.repo_id
        else:
            # Use a path restriction if we're in a repo, obviously it'll work
            # faster if we're in an invididual ebuild dir but we're not that
            # restrictive.
            try:
                restriction = namespace.repo.path_restrict(namespace.cwd)
                token = namespace.cwd
            except (AttributeError, ValueError):
                parser.error('missing target argument and not in a supported repo')

            # determine if we're grabbing the keywords for a single pkg in cwd
            namespace.pkg_dir = any(
                isinstance(x, restricts.PackageDep)
                for x in reversed(restriction.restrictions))

        namespace.targets = [(token, restriction)]


def _collapse_arches(options, pkgs):
    """Collapse arches into a single set."""
    keywords = set()
    stable_keywords = set()
    unstable_keywords = set()
    for pkg in pkgs:
        for x in pkg.keywords:
            if x[0] == '~':
                unstable_keywords.add(x[1:])
            elif x in options.arches:
                stable_keywords.add(x)
    if options.unstable:
        keywords.update(unstable_keywords)
    if options.only_unstable:
        keywords.update(unstable_keywords.difference(stable_keywords))
    if not keywords or options.stable:
        keywords.update(stable_keywords)
    return (
        sorted(keywords.intersection(options.native_arches)) +
        sorted(keywords.intersection(options.prefix_arches)))


def _render_rows(options, pkgs, arches):
    """Build rows for tabular data output."""
    for pkg in sorted(pkgs):
        keywords = set(pkg.keywords)
        row = [pkg.fullver]
        for arch in arches:
            if arch in keywords:
                line = '+'
            elif f'~{arch}' in keywords:
                line = '~'
            elif f'-{arch}' in keywords:
                line = '-'
            elif '-*' in keywords:
                line = '*'
            else:
                line = 'o'
            row.append(options.colormap(line))
        row.extend([pkg.eapi, pkg.fullslot, pkg.repo.repo_id])
        yield row


@showkw.bind_main_func
def main(options, out, err):
    continued = False
    for token, restriction in options.targets:
        for pkgs in pkgutils.groupby_pkg(options.repo.itermatch(restriction, sorter=sorted)):
            if options.collapse:
                out.write(' '.join(_collapse_arches(options, pkgs)))
            else:
                arches = sorted(options.arches.intersection(options.native_arches))
                if options.prefix:
                    arches += sorted(options.arches.intersection(options.prefix_arches))
                headers = [''] + arches + ['eapi', 'slot', 'repo']
                if continued:
                    out.write()
                if not options.pkg_dir:
                    pkgs = list(pkgs)
                    out.write(f'keywords for {pkgs[0].unversioned_atom}:')
                data = _render_rows(options, pkgs, arches)
                table = tabulate(
                    data, headers=headers, tablefmt=options.format,
                    disable_numparse=True)
                out.write(table)
            continued = True

    if not continued:
        err.write(f"{options.prog}: no matches for {token!r}")
        return 1
